import { VMWARE_OS_DICT } from 'constant/others/other';
import CloudContext from 'model/CloudContext';
import DataRecord from 'model/DataRecord';
import EntityColumn from 'model/EntityColumn';
import EntityData from 'model/EntityData';
import EntityFormTemplate from 'model/EntityFormTemplate';
import { convertDateString } from 'service/string';

/**
 * Record<string, any> to FormData.
 *
 * @param formData Record<string, any>.
 * @param entityFormTemplate Template of entity form.
 *
 * @returns FormData type.
 */
export const convertFormDataToSendData = (formData: Record<string, any>, entityFormTemplate: EntityFormTemplate) => {
  const fileTypeRecordList = [];
  for (const record1 of entityFormTemplate.entityRecords) {
    if (record1.type === 'panel') {
      for (const record2 of record1.keyValueRecords) {
        if (record2.type === 'file') {
          fileTypeRecordList.push(record2.name);
        }
      }
    }
  }

  const sendData = new FormData();
  for (const pair of Object.entries(formData)) {
    switch (typeof pair[1]) {
      case 'object':
        sendData.append(
          pair[0],
          fileTypeRecordList.includes(pair[0])
            ? pair[1]
            : JSON.stringify(pair[1])
        );
        break;
      default:
        sendData.append(pair[0], `${pair[1]}`);
        break;
    }
  }
  return sendData;
}

/**
 * Round number method.
 * @param value number
 * @param digit digit
 * @returns fixed value string
 */
const roundNumber = (value: number, digit: number) => {
  // This code means "x = 10 ^ digit".
  let x = 1.0;
  for (let i = 0; i < digit; i += 1) {
    x *= 10;
  }

  // Use guard clause.
  if (Math.floor(value) !== value) {
    return `${Math.round(value * x) / x}`;
  }

  let textBuffer = `${value}`;
  if (digit !== 0) {
    textBuffer += '.';
    for (let i = 0; i < digit; i += 1) {
      textBuffer += '0';
    }
  }
  return textBuffer;
};

const convertKeyValueCrLfData = (data: string) => {
  if (data === '{  }') {
    return '';
  }

  const records = data.split('\n');
  const CONVERT_KEY_VALUE_CRLF_DICT: Record<string, string> = {
    'on_demand_hourly: ': 'On-demand hourly ($): ',
    'on_demand_daily: ': 'On-demand daily ($): ',
    'on_demand_monthly: ': 'On-demand monthly ($): ',
    'on_demand_yearly: ': 'On-demand yearly ($): ',
    'ri_one_year: ': 'RI 1 year ($): ',
    'ri_three_year: ': 'RI 3 year ($): ',
    'cpu: ': 'CPU (Usage): ',
    'memory: !!float ': 'Memory (Usage): ',
    'memory: ': 'Memory (Usage): ',
    'pod_allocation: ': 'Pods (Allocation): ',
    'cpu_capacity: !!float ': 'CPU (Capacity): ',
    'memory_capacity: !!float ': 'Memory (Capacity): ',
    'pod_capacity: ': 'Pods (Capacity): ',
  };
  const fixedRecords: string[] = [];
  for (const record of records) {
    if (record === '') {
      continue;
    }
    let fixedRecord = record;
    for (const key in CONVERT_KEY_VALUE_CRLF_DICT) {
      if (record.includes(key)) {
        const value = record.replace(key, '');
        const valueFloat = parseFloat(value);
        if (key.includes('!!float') || key.includes('memory: ')) {
          if (valueFloat >= 1024 * 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024 * 1024), 2)}Gi`;
          } else if (valueFloat >= 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024), 2)}Mi`;
          } else if (valueFloat >= 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024), 2)}Ki`;
          } else {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat, 2)}`;
          }
        } else {
          fixedRecord = CONVERT_KEY_VALUE_CRLF_DICT[key].includes('Pods ')
            ? CONVERT_KEY_VALUE_CRLF_DICT[key] + value
            : CONVERT_KEY_VALUE_CRLF_DICT[key] + roundNumber(valueFloat, 2);
        }
        break;
      }
    }
    fixedRecords.push(fixedRecord);
  }
  return fixedRecords.sort((a, b) => {
    return a < b ? -1 : a > b ? 1 : 0;
  }).join('\n');
}

/**
 * Converter of entity's data for UI.
 * @param data Entity's data
 * @param ec Information of Entity Column
 * @param dataCache Data cache for 'join' type
 */
export const convertDataToTextForUI = (data: any, ec: EntityColumn, dataCache: { [key: string]: EntityData[] }): string => {
  // Null check.
  if (data === null) {
    return '';
  }

  switch (ec.type) {
    case 'datetime':
      return convertDateString(data);
    case 'cpu':
      return roundNumber(data, 2);
    case 'memory': {
      if (data >= 1024 * 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024 * 1024), 2)}Gi`;
      }
      if (data >= 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024), 2)}Mi`;
      }
      if (data >= 1024) {
        return `${roundNumber(data / (1024), 2)}Ki`;
      }
      return roundNumber(data, 2);
    }
    case 'key-value': {
      let records: string[] = [];
      for (const record of data) {
        records.push(`${record['item_key']}:${record['item_value']}`);
      }
      return records.join(', ');
    }
    case 'cost':
      return `$${data}`;
    case 'boolean':
      return data ? ec.value[0] : ec.value[1];
    case 'array':
      if (!Array.isArray(data)) {
        return data;
      }
      return data.map((r: any) => `${r}`).join(', ');
    case 'join':
      for (const record of dataCache[ec.info.entityTypeId]) {
        const recordData: EntityData = record;
        if (recordData.attributes[ec.info.keyColumn] === data) {
          return recordData.attributes[ec.info.valueColumn];
        }
      }
      return `${data}`;
    case 'key-value-crlf':
      return convertKeyValueCrLfData(data);
    case 'conditions':
      return data !== ec.value[0] ? data : ec.value[1];
    case 'cidr':
      return data.length > 0 ? data[0]['cidr'] : '';
    case 'number':
      return data.toString();
    case 'vmware_os':
      // source: VmwareVmGuestOsDataProvider.php
      return data in VMWARE_OS_DICT ? VMWARE_OS_DICT[data] : data;
    case 'link':
      return JSON.stringify({
        'url': ec.url,
        'label': ec.label,
        'type': ec.type,
      });
    default:
      return data;
  }
};

/**
 * Convert EntityData to DataRecord.
 *
 * @param rawDataList List of EntityData.
 * @param entityColumnList List of EntityColumn.
 * @param cloudContext cloud_context.
 * @param dataCache Data cache of EntityData.
 * @returns List of DataRecord.
 */
export const convertEntityDataToDataRecord = (
  entityTypeId: string,
  rawDataList: EntityData[],
  entityColumnList: EntityColumn[],
  cloudContextList: CloudContext[],
  dataCache: { [key: string]: EntityData[] }) => {
  const newDataRecordList: DataRecord[] = [];
  for (const rawRecord of rawDataList) {
    // convert (rawRecord: EntityData) => (dataRecord: DataRecord)
    const dataRecord: DataRecord = {
      id: `${rawRecord.attributes['drupal_internal__id']}`,
      entityTypeId,
      value: {},
      cloudContext: ''
    };
    dataRecord.cloudContext = rawRecord.attributes['cloud_context'];

    // Read data column by launchTemplateColumn, convert as appropriate, and write it out.
    for (const launchTemplateColumn of entityColumnList) {
      if (launchTemplateColumn.name in rawRecord.attributes) {
        // Read data column by launchTemplateColumn.
        const rawValue = rawRecord.attributes[launchTemplateColumn.name];

        // Convert as appropriate.
        if (launchTemplateColumn.name !== 'cloud_context') {
          // If it is not 'cloud_context', convert it by convertDataToTextForUI().
          dataRecord.value[launchTemplateColumn.name] = convertDataToTextForUI(rawValue, launchTemplateColumn, dataCache);
          continue;
        }
        // Search for possible candidates.
        const candidate = cloudContextList.filter((r) => {
          return r.name !== 'ALL' && r.name === rawValue;
        });
        // Branching depending on the number of candidates.
        dataRecord.value[launchTemplateColumn.name] = candidate.length >= 1
          ? candidate[0].labelName
          : rawValue;
        continue;
      }

      if (launchTemplateColumn.type === 'relationship' && typeof rawRecord.relationships !== 'string') {
        if (launchTemplateColumn.name in rawRecord.relationships) {
          // Read data column by launchTemplateColumn.
          const rawValue = rawRecord.relationships[launchTemplateColumn.name];
          const relationshipsData = Array.isArray(rawValue.data) ? rawValue.data[0] : rawValue.data;

          const entityId = relationshipsData.meta[launchTemplateColumn.info.keyColumn1];
          const entityList = dataCache[launchTemplateColumn.info.entityTypeId];
          const selectedEntity = entityList.filter((e) => {
            return e.attributes[launchTemplateColumn.info.keyColumn2] === entityId;
          });
          dataRecord.value[launchTemplateColumn.name] = selectedEntity.length >= 1
            ? selectedEntity[0].attributes[launchTemplateColumn.info.valueColumn]
            : '';
        }
      }
    }
    // Write it out.
    newDataRecordList.push(dataRecord);
  }
  return newDataRecordList;
};
